const { getLocalizedMessage } = require("../lib/locale-provider");

const { handleError } = require("../lib/notifications");
const { handleLogin, singleOidcSetup, oidcSetup } = require("../lib/login");
const { wrapUI } = require("./components/wrap");
const { inputWithLabel } = require("./components/inputs");
const { primaryButtonUI, navButtonUI, oidcButtonUI } = require("./components/buttons");
const { text } = require("./components/texts");
const { divider } = require("./components/divider");
const { getElementIds } = require("../lib/config");
const { getNotificationsState } = require("../lib/notifications");
const { treatAsTypeLogical, treatAsTypeOidc } = require("../utils/helpers");

const elementIds = getElementIds();
const loginUI = ({
  context,
  labels,
  oidc,
  redirectUri,
  onSuccessCallback,
  registrationPath = "/registration",
  forgotPasswordPath = "/forgot/password",
}) => {
  const defaultLabels = {
    username: getLocalizedMessage("uisdk.general.username"),
    password: getLocalizedMessage("uisdk.general.password"),
    loginButton: getLocalizedMessage("uisdk.login.login"),
    registerButton: getLocalizedMessage("uisdk.login.register_button"),
    forgotPasswordButton: getLocalizedMessage("uisdk.login.forgot_password_button"),
    orOtherOptions: getLocalizedMessage("uisdk.login.other_options"),
  };

  // We are detecting whether opts exist in the top layer as server message can vary based on @type
  const isTypeLogical = treatAsTypeLogical(context);
  const isTypeOidc = treatAsTypeOidc(context);

  const actions = isTypeLogical && context.opts.find((option) => option["@type"] === "action");
  const forgottenAction =
    actions &&
    actions.opts &&
    Array.isArray(actions.opts) &&
    actions.opts.find((opt) => opt.hint === "forgotten");

  const registerAction =
    actions &&
    actions.opts &&
    Array.isArray(actions.opts) &&
    actions.opts.find((opt) => opt.hint === "register");

  const loginFormFrame = document.createElement("div");
  loginFormFrame.className = "login-form";

  // @type is oidc
  if (isTypeOidc) {
    loginFormFrame.appendChild(
      oidcButtonUI({
        data: context,
        onClick: () => oidcSetup(null, redirectUri),
      }),
    );
  }

  // @type is logical and opts include form
  if (isTypeLogical && context.opts.some((i) => i["@type"] === "form")) {
    const form = document.createElement("form");
    const { label: userLabel, input: userInput } = inputWithLabel({
      type: "email",
      id: elementIds.inputEmail,
      labelText: (labels && labels.username) || defaultLabels.username,
      autofocus: true,
    });
    const { label: psswrdLabel, input: psswrdInput } = inputWithLabel({
      type: "password",
      id: elementIds.inputPassword,
      labelText: (labels && labels.password) || defaultLabels.password,
    });
    form.appendChild(userLabel);
    form.appendChild(userInput);
    form.appendChild(psswrdLabel);
    form.appendChild(psswrdInput);
    form.appendChild(
      primaryButtonUI({
        id: elementIds.submitLoginBtn,
        onClick: (e) => {
          e.preventDefault();
          handleLogin({
            id: isTypeLogical
              ? context.opts.filter((o) => o["@type"] === "form")[0]["@id"]
              : undefined,
            oidc,
            onSuccessCallback,
            emailValueParam: null,
            passwordValueParam: null,
          }).catch((err) => {
            console.log(err);
            if (typeof err === "object" && err["~error"]) {
              handleError(err["~error"]);
            }
          });
        },
        label: (labels && labels.loginButton) || defaultLabels.loginButton,
      }),
    );

    loginFormFrame.appendChild(form);
  }

  if (!isTypeOidc) {
    // If @type oidc we want to handle all authorization via the identity provider so no need for this.
    const hr = divider();
    loginFormFrame.appendChild(hr);

    if (registerAction) {
      const createNewAccountBtn = navButtonUI({
        id: elementIds.registrationBtn,
        label: (labels && labels.registerButton) || defaultLabels.registerButton,
        href: registrationPath,
      });
      loginFormFrame.appendChild(createNewAccountBtn);
    }
  }

  if (forgottenAction) {
    const forgotPasswordBtn = navButtonUI({
      id: elementIds.forgotPasswordBtn,
      label: (labels && labels.forgotPasswordButton) || defaultLabels.forgotPasswordButton,
      href: forgotPasswordPath + document.location.search,
    });
    loginFormFrame.appendChild(forgotPasswordBtn);
  }

  const secondHr = divider();
  loginFormFrame.appendChild(secondHr);

  // @type is logical and opts includes oidc options
  if (isTypeLogical && context.opts.some((o) => o["@type"] === "oidc")) {
    const otherMethodsContainerEl = document.createElement("div");
    otherMethodsContainerEl.className = "other-methods-container";
    const otherMethodsHintEl = text(
      (labels && labels.orOtherOptions) || defaultLabels.orOtherOptions,
    );
    otherMethodsContainerEl.appendChild(otherMethodsHintEl);
    context.opts
      .filter((o) => o["@type"] === "oidc")
      .forEach((o) => {
        otherMethodsContainerEl.appendChild(
          oidcButtonUI({
            id: elementIds.oidcBtnPrefix + o.prv.replace(".com", "").toLowerCase(),
            data: o,
            onClick: () => oidcSetup(o["@id"], redirectUri),
          }),
        );
      });

    loginFormFrame.appendChild(otherMethodsContainerEl);
  }

  const notificationState = getNotificationsState();
  return wrapUI(loginFormFrame, notificationState);
};

module.exports = loginUI;
