const { getLocalizedMessage } = require("../lib/locale-provider");
const { handleError } = require("../lib/notifications");
const { wrapUI } = require("./components/wrap");
const { getElementIds } = require("../lib/config");
const { oidcSetup } = require("../lib/login");
const { inputWithLabel } = require("./components/inputs");
const { divider } = require("./components/divider");
const { text } = require("./components/texts");
const { primaryButtonUI, navButtonUI, oidcButtonUI } = require("./components/buttons");
const { handleRegister } = require("../lib/register");
const { treatAsTypeLogical } = require("../utils/helpers");

const elementIds = getElementIds();

const registerUI = ({
  context,
  labels,
  loginPath = "/login",
  redirectUri,
  onSuccessCallback,
  validatePassword,
  termsAngAgreementHtmlString,
  oidc,
  userInputNote,
  passwordInputNote,
  passwordCheckInputNote,
}) => {
  const defaultLabels = {
    username: getLocalizedMessage("uisdk.general.username"),
    password: getLocalizedMessage("uisdk.general.password"),
    confirmPassword: getLocalizedMessage("uisdk.register.confirm_password"),
    registerButton: getLocalizedMessage("uisdk.register.register_button"),
    alreadyHaveAnAccountButton: getLocalizedMessage("uisdk.register.have_account_button"),
    orOtherOptions: getLocalizedMessage("uisdk.register.other_options"),
  };

  const registrationFormFrame = document.createElement("div");
  registrationFormFrame.className = "registration-form";
  // We are detecting whether opts exist in the top layer as server message can vary based on @type
  const hasOptsField = treatAsTypeLogical(context);

  const form = document.createElement("form");
  const { label: userLabel, input: userInput, note: userNote } = inputWithLabel({
    type: "email",
    id: elementIds.inputEmail,
    labelText: (labels && labels.username) || defaultLabels.username,
    autofocus: true,
    noteText: userInputNote,
  });

  const { label: psswrdLabel, input: psswrdInput, note: psswrdNote } = inputWithLabel({
    type: "password",
    id: elementIds.inputPassword,
    labelText: (labels && labels.password) || defaultLabels.password,
    noteText: passwordInputNote,
  });

  const {
    label: psswrdConfirmLabel,
    input: psswrdConfirmInput,
    note: psswrdConfirmNote,
  } = inputWithLabel({
    type: "password",
    id: elementIds.inputConfirmPassword,
    labelText: (labels && labels.confirmPassword) || defaultLabels.confirmPassword,
    noteText: passwordCheckInputNote,
  });

  form.appendChild(userLabel);
  form.appendChild(userInput);
  if (userNote) {
    form.appendChild(userNote);
  }
  form.appendChild(psswrdLabel);
  form.appendChild(psswrdInput);
  if (psswrdNote) {
    form.appendChild(psswrdNote);
  }
  form.appendChild(psswrdConfirmLabel);
  form.appendChild(psswrdConfirmInput);
  if (psswrdConfirmNote) {
    form.appendChild(psswrdConfirmNote);
  }
  form.appendChild(
    primaryButtonUI({
      id: elementIds.submitRegisterBtn,
      onClick: (e) => {
        e.preventDefault();
        // check if property validatePassword exists and is a function
        if (typeof validatePassword === "function") {
          if (validatePassword(psswrdInput.value) !== true) {
            //abort if validatePassword not returns true
            // IndyRiot doesn't want to have errors
            // handleError({
            //   label: "uisdk.general.password_validation",
            // });
            return;
          }
        }
        handleRegister({
          id: hasOptsField
            ? context.opts.filter((o) => o["@type"] === "form")[0]["@id"]
            : undefined,
          oidc,
          onSuccessCallback,
          emailValueParam: null,
          passwordValueParam: null,
        }).catch((err) => {
          if (typeof err === "object" && err["~error"]) {
            handleError(err["~error"]);
          }
        });
      },
      label:
        (labels && labels.registerButton) ||
        (termsAngAgreementHtmlString
          ? `Agree & ${defaultLabels.registerButton}`
          : defaultLabels.registerButton),
    }),
  );
  registrationFormFrame.appendChild(form);

  if (termsAngAgreementHtmlString) {
    const termsAngAgreementContainer = document.createElement("div");
    termsAngAgreementContainer.style.cssText = "text-align: center;";
    termsAngAgreementContainer.innerHTML = termsAngAgreementHtmlString;
    registrationFormFrame.appendChild(termsAngAgreementContainer);
  }

  const hr = divider();
  registrationFormFrame.appendChild(hr);

  const alreadyRegisteredNavBtn = navButtonUI({
    id: elementIds.loginBtn,
    label:
      (labels && labels.alreadyHaveAnAccountButton) || defaultLabels.alreadyHaveAnAccountButton,
    href: loginPath + document.location.search,
  });
  registrationFormFrame.appendChild(alreadyRegisteredNavBtn);

  const secondHr = divider();
  registrationFormFrame.appendChild(secondHr);

  if (hasOptsField && context.opts.some((o) => o["@type"] === "oidc")) {
    const otherMethodsContainerEl = document.createElement("div");
    otherMethodsContainerEl.className = "other-methods-container";
    const otherMethodsHintEl = text(
      (labels && labels.orOtherOptions) || defaultLabels.orOtherOptions,
    );
    otherMethodsContainerEl.appendChild(otherMethodsHintEl);
    context.opts
      .filter((o) => o["@type"] === "oidc")
      .forEach((o) => {
        otherMethodsContainerEl.appendChild(
          oidcButtonUI({
            id: elementIds.oidcBtnPrefix + o.prv.replace(".com", "").toLowerCase(),
            data: o,
            onClick: () => oidcSetup(o["@id"], redirectUri),
          }),
        );
      });

    registrationFormFrame.appendChild(otherMethodsContainerEl);
  }

  return wrapUI(registrationFormFrame);
};

module.exports = registerUI;
